#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/time.h>
#include <ctype.h>

#include "AVAPIs_Server.h"
#include "speaker/speaker.h"
#include "playback/playback.h"
#include "emulator/emulator.h"
#include "ability/ability.h"
#include "token/token.h"
#include "dasa/av_dasa.h"
#include "NebulaError.h"
#include "PushNotification.h"
#include "sample_constants.h"


static char gAvAccount[NEW_MAXSIZE_VIEWACC] = "admin";
static char gAvPassword[NEW_MAXSIZE_VIEWPWD] = "888888";
static char gIotcAuthkey[IOTC_AUTH_KEY_LENGTH+1] = "00000000"; //use "00000000" as default authkey

AV_Client gClientInfo[MAX_CLIENT_NUMBER];

static char gAbilityFile[128] = "";
static char gDefaultSettingsFilePath[128] = "./notification_setting.txt";
char gVideoFn[128] = "./video/I_000.bin";
char gAudioFn[128] = "./audio/8k_16bit_mono.raw";
int gProcessRun = 1;
int gOnlineNum = 0;
bool gEnablePushNotification = false;
bool gSettingsExist = false;

static char gUID[MAX_UID_LENGTH+1] = {0};

static void PrintErrHandling (int nErr)
{
	switch (nErr)
	{
	case IOTC_ER_MASTER_NOT_RESPONSE :
		//-60 IOTC_ER_MASTER_NOT_RESPONSE
		printf ("[Error code : %d]\n", IOTC_ER_MASTER_NOT_RESPONSE );
		printf ("Master server doesn't respond.\n");
		printf ("Please check the network wheather it could connect to the Internet.\n");
		break;
	case IOTC_ER_SERVER_NOT_RESPONSE :
		//-1 IOTC_ER_SERVER_NOT_RESPONSE
		printf ("[Error code : %d]\n", IOTC_ER_SERVER_NOT_RESPONSE );
		printf ("P2P Server doesn't respond.\n");
		printf ("Please check the network wheather it could connect to the Internet.\n");
		break;
	case IOTC_ER_FAIL_RESOLVE_HOSTNAME :
		//-2 IOTC_ER_FAIL_RESOLVE_HOSTNAME
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_RESOLVE_HOSTNAME);
		printf ("Can't resolve hostname.\n");
		break;
	case IOTC_ER_ALREADY_INITIALIZED :
		//-3 IOTC_ER_ALREADY_INITIALIZED
		printf ("[Error code : %d]\n", IOTC_ER_ALREADY_INITIALIZED);
		printf ("Already initialized.\n");
		break;
	case IOTC_ER_FAIL_CREATE_MUTEX :
		//-4 IOTC_ER_FAIL_CREATE_MUTEX
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_MUTEX);
		printf ("Can't create mutex.\n");
		break;
	case IOTC_ER_FAIL_CREATE_THREAD :
		//-5 IOTC_ER_FAIL_CREATE_THREAD
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_CREATE_THREAD);
		printf ("Can't create thread.\n");
		break;
	case IOTC_ER_UNLICENSE :
		//-10 IOTC_ER_UNLICENSE
		printf ("[Error code : %d]\n", IOTC_ER_UNLICENSE);
		printf ("This UID is unlicense.\n");
		printf ("Check your UID.\n");
		break;
	case IOTC_ER_NOT_INITIALIZED :
		//-12 IOTC_ER_NOT_INITIALIZED
		printf ("[Error code : %d]\n", IOTC_ER_NOT_INITIALIZED);
		printf ("Please initialize the IOTCAPI first.\n");
		break;
	case IOTC_ER_TIMEOUT :
		//-13 IOTC_ER_TIMEOUT
		break;
	case IOTC_ER_INVALID_SID :
		//-14 IOTC_ER_INVALID_SID
		printf ("[Error code : %d]\n", IOTC_ER_INVALID_SID);
		printf ("This SID is invalid.\n");
		printf ("Please check it again.\n");
		break;
	case IOTC_ER_EXCEED_MAX_SESSION :
		//-18 IOTC_ER_EXCEED_MAX_SESSION
		printf ("[Error code : %d]\n", IOTC_ER_EXCEED_MAX_SESSION);
		printf ("[Warning]\n");
		printf ("The amount of session reach to the maximum.\n");
		printf ("It cannot be connected unless the session is released.\n");
		break;
	case IOTC_ER_CAN_NOT_FIND_DEVICE :
		//-19 IOTC_ER_CAN_NOT_FIND_DEVICE
		printf ("[Error code : %d]\n", IOTC_ER_CAN_NOT_FIND_DEVICE);
		printf ("Device didn't register on server, so we can't find device.\n");
		printf ("Please check the device again.\n");
		printf ("Retry...\n");
		break;
	case IOTC_ER_SESSION_CLOSE_BY_REMOTE :
		//-22 IOTC_ER_SESSION_CLOSE_BY_REMOTE
		printf ("[Error code : %d]\n", IOTC_ER_SESSION_CLOSE_BY_REMOTE);
		printf ("Session is closed by remote so we can't access.\n");
		printf ("Please close it or establish session again.\n");
		break;
	case IOTC_ER_REMOTE_TIMEOUT_DISCONNECT :
		//-23 IOTC_ER_REMOTE_TIMEOUT_DISCONNECT
		printf ("[Error code : %d]\n", IOTC_ER_REMOTE_TIMEOUT_DISCONNECT);
		printf ("We can't receive an acknowledgement character within a TIMEOUT.\n");
		printf ("It might that the session is disconnected by remote.\n");
		printf ("Please check the network wheather it is busy or not.\n");
		printf ("And check the device and user equipment work well.\n");
		break;
	case IOTC_ER_DEVICE_NOT_LISTENING :
		//-24 IOTC_ER_DEVICE_NOT_LISTENING
		printf ("[Error code : %d]\n", IOTC_ER_DEVICE_NOT_LISTENING);
		printf ("Device doesn't listen or the sessions of device reach to maximum.\n");
		printf ("Please release the session and check the device wheather it listen or not.\n");
		break;
	case IOTC_ER_CH_NOT_ON :
		//-26 IOTC_ER_CH_NOT_ON
		printf ("[Error code : %d]\n", IOTC_ER_CH_NOT_ON);
		printf ("Channel isn't on.\n");
		printf ("Please open it by IOTC_Session_Channel_ON() or IOTC_Session_Get_Free_Channel()\n");
		printf ("Retry...\n");
		break;
	case IOTC_ER_SESSION_NO_FREE_CHANNEL :
		//-31 IOTC_ER_SESSION_NO_FREE_CHANNEL
		printf ("[Error code : %d]\n", IOTC_ER_SESSION_NO_FREE_CHANNEL);
		printf ("All channels are occupied.\n");
		printf ("Please release some channel.\n");
		break;
	case IOTC_ER_TCP_TRAVEL_FAILED :
		//-32 IOTC_ER_TCP_TRAVEL_FAILED
		printf ("[Error code : %d]\n", IOTC_ER_TCP_TRAVEL_FAILED);
		printf ("Device can't connect to Master.\n");
		printf ("Don't let device use proxy.\n");
		printf ("Close firewall of device.\n");
		printf ("Or open device's TCP port 80, 443, 8080, 8000, 21047.\n");
		break;
	case IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED :
		//-33 IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED
		printf ("[Error code : %d]\n", IOTC_ER_TCP_CONNECT_TO_SERVER_FAILED);
		printf ("Device can't connect to server by TCP.\n");
		printf ("Don't let server use proxy.\n");
		printf ("Close firewall of server.\n");
		printf ("Or open server's TCP port 80, 443, 8080, 8000, 21047.\n");
		printf ("Retry...\n");
		break;
	case IOTC_ER_NO_PERMISSION :
		//-40 IOTC_ER_NO_PERMISSION
		printf ("[Error code : %d]\n", IOTC_ER_NO_PERMISSION);
		printf ("This UID's license doesn't support TCP.\n");
		break;
	case IOTC_ER_NETWORK_UNREACHABLE :
		//-41 IOTC_ER_NETWORK_UNREACHABLE
		printf ("[Error code : %d]\n", IOTC_ER_NETWORK_UNREACHABLE);
		printf ("Network is unreachable.\n");
		printf ("Please check your network.\n");
		printf ("Retry...\n");
		break;
	case IOTC_ER_FAIL_SETUP_RELAY :
		//-42 IOTC_ER_FAIL_SETUP_RELAY
		printf ("[Error code : %d]\n", IOTC_ER_FAIL_SETUP_RELAY);
		printf ("Client can't connect to a device via Lan, P2P, and Relay mode\n");
		break;
	case IOTC_ER_NOT_SUPPORT_RELAY :
		//-43 IOTC_ER_NOT_SUPPORT_RELAY
		printf ("[Error code : %d]\n", IOTC_ER_NOT_SUPPORT_RELAY);
		printf ("Server doesn't support UDP relay mode.\n");
		printf ("So client can't use UDP relay to connect to a device.\n");
		break;

	default :
		printf("[Error code : %d]\n", nErr);
		break;
	}
}

static void printAccount()
{
	printf("=====Account_info=======\n");
	printf("gAvAccount:%s\n", gAvAccount);
	printf("gAvPassword:%s\n", gAvPassword);
	printf("gIotcAuthkey:%s\n", gIotcAuthkey);
	printf("========================\n");
}


#if ENABLE_TOKEN_AUTH
// Advance use of authentication.
// Users can enable or disable these function depends on the actual situation.
static int ExTokenAuthCallBackFn(const char *identity, char *token, unsigned int token_buf_size){
    char token_fetch[TOKEN_LENGTH+1];
    unsigned int token_fetch_length;

    // fetch token
    if(GetToken(identity, token_fetch, &token_fetch_length)!=TOKEN_SUCCESS)
        return -1;

    // store token into buffer and it will be checked
    // after this callback function
    if (token_buf_size <= token_fetch_length) return -1;
    strcpy(token, token_fetch);

    return 0;
}

static int ExTokenDeleteCallBackFn(int av_index, const char *identity) {
    int ret = 0;
    ret = DeleteTokenByIdentity(identity, strlen(identity));
    return ret;
}

static int ExTokenRequestCallBackFn(
        int av_index,
        const char *identity,
        const char *identity_description,
        char *token,
        unsigned int token_length) {
    int ret = 0;
    TokenPrivilege privilege = GUEST_PRIVILEGE;

    if(identity_description != NULL && strcmp(identity_description, "admin"))
        privilege = ADMIN_PRIVILEGE;

    ret = CreateTokenWithIdentity(identity, strlen(identity), TOKEN_LENGTH, token, privilege);

    return ret;
}

static void ExGetIdentityArrayCallBackFn(int avIndex, avServSendIdentityArray send_identity_array) {
    int ret = 0;
    AvIdentity *identities = NULL;
    unsigned int identity_count;

    ret = GetIdentityArray(&identities, &identity_count);
    if (ret < 0) {
        send_identity_array(avIndex, NULL, 0, ret);
    } else {
        send_identity_array(avIndex, identities, identity_count, ret);
        FreeIdentityArray(identities);
    }
}
#endif

static int NotificationSettingHandleFn(const char *setting_str)
{
    printf("NotificationSettingHandleFn\n");
    printf("%s\n", setting_str);

    FILE *fp = fopen(gDefaultSettingsFilePath, "w+");
    if (fp) {
        fwrite(setting_str, 1, strlen(setting_str), fp);
        fclose(fp);
        printf("[%s] get new settings, save to %s success\n", __func__, gDefaultSettingsFilePath);
    } else {
        printf("[%s] get new settings, save to %s fail\n", __func__, gDefaultSettingsFilePath);
    }
    gSettingsExist = true;
    return 0;
}

int ExPasswordAuthCallBackFn(const char *account, char *pwd, unsigned int pwd_buf_size)
{
	if (strcmp(account, gAvAccount) != 0){
		return -1;
    }
    if (pwd_buf_size <= strlen(gAvPassword)){
        return -1;
    }

	strcpy(pwd, gAvPassword);

	return 0;
}

static int ExChangePasswordCallBackFn(
        int av_index,
        const char *account,
        const char *old_password,
        const char *new_password,
        const char *new_iotc_authkey){

	if (strcmp(account, gAvAccount) != 0)
		return -1;
	if (strcmp(old_password, gAvPassword) != 0)
		return -1;

    // please make sure the maximum password size is enough
	if (strlen(new_password) > NEW_MAXSIZE_VIEWPWD)
		return -1;
	strcpy(gAvPassword, new_password);

	//auth key has been changed inside the SDK, please save it
    memcpy(gIotcAuthkey, new_iotc_authkey, IOTC_AUTH_KEY_LENGTH);

    IOTC_Device_Update_Authkey(gIotcAuthkey);
	printAccount();
	return 0; //success
}

static void ExAbilityRequestFn(int av_index, avServSendAbility send_ability)
{
    char *buf = NULL;

    if (readAbility(gAbilityFile, &buf) > 0) {
        send_ability(av_index, (const unsigned char*)buf, (unsigned int)strlen(buf)+1);
        releaseAbility(&buf);
    } else {
        send_ability(av_index, (const unsigned char*)"NULL", 4);
    }
}

static void regedit_client_to_video(int SID, int avIndex)
{
	AV_Client *p = &gClientInfo[SID];
	p->avIndex = avIndex;
	p->bEnableVideo = 1;
	p->waitKeyFrame = 1;
}

static void unregedit_client_from_video(int SID)
{
	AV_Client *p = &gClientInfo[SID];
	p->bEnableVideo = 0;
	p->speakerThreadId = 0;
}

static void regedit_client_to_audio(int SID, int avIndex)
{
	AV_Client *p = &gClientInfo[SID];
	p->bEnableAudio = 1;
}

static void unregedit_client_from_audio(int SID)
{
	AV_Client *p = &gClientInfo[SID];
	p->bEnableAudio = 0;
}

static void regedit_client_stream_mode(int SID, int streamMode)
{
    AV_Client *p = &gClientInfo[SID];
	p->bTwoWayStream = streamMode;
}

static void Handle_IOCTRL_Cmd(int SID, int avIndex, char *buf, int type)
{
	printf("Handle CMD: ");
	switch(type)
	{
	case IOTYPE_USER_IPCAM_START:
		{
			SMsgAVIoctrlAVStream *p = (SMsgAVIoctrlAVStream *)buf;
			printf("IOTYPE_USER_IPCAM_START, ch:%d, avIndex:%d\n\n", p->channel, avIndex);
			//get writer lock
			int lock_ret = pthread_rwlock_wrlock(&gClientInfo[SID].sLock);
			if(lock_ret)
				printf("Acquire SID %d rwlock error, ret = %d\n", SID, lock_ret);
			regedit_client_to_video(SID, avIndex);
			//release lock
			lock_ret = pthread_rwlock_unlock(&gClientInfo[SID].sLock);
			if(lock_ret)
				printf("Release SID %d rwlock error, ret = %d\n", SID, lock_ret);

			printf("regedit_client_to_video OK\n");

#if ENABLE_DASA
			avDASAReset(avIndex, AV_DASA_LEVEL_QUALITY_LOW);
#endif

#if ENABLE_EMULATOR
            Emulator_ForceSendIFrame();
#endif
		}
		break;
	case IOTYPE_USER_IPCAM_STOP:
		{
			SMsgAVIoctrlAVStream *p = (SMsgAVIoctrlAVStream *)buf;
			printf("IOTYPE_USER_IPCAM_STOP, ch:%d, avIndex:%d\n\n", p->channel, avIndex);
			//get writer lock
			int lock_ret = pthread_rwlock_wrlock(&gClientInfo[SID].sLock);
			if(lock_ret)
				printf("Acquire SID %d rwlock error, ret = %d\n", SID, lock_ret);
			unregedit_client_from_video(SID);
			//release lock
			lock_ret = pthread_rwlock_unlock(&gClientInfo[SID].sLock);
			if(lock_ret)
				printf("Release SID %d rwlock error, ret = %d\n", SID, lock_ret);
			printf("unregedit_client_from_video OK\n");
		}
		break;
	case IOTYPE_USER_IPCAM_AUDIOSTART:
		{
			SMsgAVIoctrlAVStream *p = (SMsgAVIoctrlAVStream *)buf;
			printf("IOTYPE_USER_IPCAM_AUDIOSTART, ch:%d, avIndex:%d\n\n", p->channel, avIndex);
			//get writer lock
			int lock_ret = pthread_rwlock_wrlock(&gClientInfo[SID].sLock);
			if(lock_ret)
				printf("Acquire SID %d rwlock error, ret = %d\n", SID, lock_ret);
			regedit_client_to_audio(SID, avIndex);
			//release lock
			lock_ret = pthread_rwlock_unlock(&gClientInfo[SID].sLock);
			if(lock_ret)
				printf("Release SID %d rwlock error, ret = %d\n", SID, lock_ret);
			printf("regedit_client_to_audio OK\n");
		}
		break;
	case IOTYPE_USER_IPCAM_AUDIOSTOP:
		{
			SMsgAVIoctrlAVStream *p = (SMsgAVIoctrlAVStream *)buf;
			printf("IOTYPE_USER_IPCAM_AUDIOSTOP, ch:%d, avIndex:%d\n\n", p->channel, avIndex);
			//get writer lock
			int lock_ret = pthread_rwlock_wrlock(&gClientInfo[SID].sLock);
			if(lock_ret)
				printf("Acquire SID %d rwlock error, ret = %d\n", SID, lock_ret);
			unregedit_client_from_audio(SID);
			//release lock
			lock_ret = pthread_rwlock_unlock(&gClientInfo[SID].sLock);
			if(lock_ret)
				printf("Release SID %d rwlock error, ret = %d\n", SID, lock_ret);
			printf("unregedit_client_from_audio OK\n");
		}
		break;
		case IOTYPE_USER_IPCAM_GETAUDIOOUTFORMAT_REQ:
		{
            if (((SMsgAVIoctrlGetAudioOutFormatReq *)buf)->channel != 0)
                gClientInfo[SID].bTwoWayStream = 0;

			Handle_IOTYPE_USER_IPCAM_GETAUDIOOUTFORMAT_REQ(avIndex, buf);
		}
		break;
		case IOTYPE_USER_IPCAM_SPEAKERSTART:
		{
			Handle_IOTYPE_USER_IPCAM_SPEAKERSTART(SID, avIndex, buf);
		}
		break;
	case IOTYPE_USER_IPCAM_SPEAKERSTOP:
		{
			Handle_IOTYPE_USER_IPCAM_SPEAKERSTOP(SID, avIndex, buf);
		}
		break;
	case IOTYPE_USER_IPCAM_LISTEVENT_REQ:
		{
			Handle_IOTYPE_USER_IPCAM_LISTEVENT_REQ(avIndex, buf);
		}
		break;
	case IOTYPE_USER_IPCAM_RECORD_PLAYCONTROL:
		{
			Handle_IOTYPE_USER_IPCAM_RECORD_PLAYCONTROL(SID, avIndex, buf);
		}
		break;
	default:
		printf("avIndex %d: non-handle type[%X]\n", avIndex, type);
		break;
	}
}

/****
Thread - Start AV server and recv IOCtrl cmd for every new av idx
*/
static void *thread_ForAVServerStart(void *arg)
{
	int SID = *(int *)arg;
	free(arg);
	int ret;
	unsigned int ioType;
	char ioCtrlBuf[AV_MAX_IOCTRL_DATA_SIZE];
	struct st_SInfoEx Sinfo;

	printf("SID[%d], thread_ForAVServerStart, OK\n", SID);

	AVServStartInConfig avStartInConfig;
    AVServStartOutConfig avStartOutConfig;

    memset(&avStartInConfig, 0, sizeof(AVServStartInConfig));
    avStartInConfig.cb               = sizeof(AVServStartInConfig);
    avStartInConfig.iotc_session_id  = SID;
    avStartInConfig.iotc_channel_id  = 0;
    avStartInConfig.timeout_sec      = 30;
    avStartInConfig.password_auth    = ExPasswordAuthCallBackFn;
    avStartInConfig.server_type      = SERVTYPE_STREAM_SERVER;
    avStartInConfig.resend           = ENABLE_RESEND;
    avStartInConfig.change_password_request = ExChangePasswordCallBackFn;
    avStartInConfig.ability_request  = ExAbilityRequestFn;
#if ENABLE_TOKEN_AUTH
// Advance use of authentication.
// Users can enable or disable these function depends on the actual situation.
    avStartInConfig.token_auth       = ExTokenAuthCallBackFn;
    avStartInConfig.token_delete     = ExTokenDeleteCallBackFn;
    avStartInConfig.token_request    = ExTokenRequestCallBackFn;
    avStartInConfig.identity_array_request = ExGetIdentityArrayCallBackFn;
#endif
    
#if ENABLE_DTLS
// Enable DTLS encryption of AV data, otherwise use AV_SECURITY_SIMPLE
    avStartInConfig.security_mode = AV_SECURITY_DTLS; 
#else
    avStartInConfig.security_mode = AV_SECURITY_SIMPLE;
#endif

    avStartOutConfig.cb              = sizeof(AVServStartOutConfig);

    int avIndex = avServStartEx(&avStartInConfig, &avStartOutConfig);

    if(avIndex < 0)
	{
		printf("avServStartEx failed!! SID[%d] code[%d]!!!\n", SID, avIndex);
		printf("thread_ForAVServerStart: exit!! SID[%d]\n", SID);
		IOTC_Session_Close(SID);
		gOnlineNum--;
		pthread_exit(0);
	}
    Sinfo.size = sizeof(struct st_SInfoEx);
	if(IOTC_Session_Check_Ex(SID, &Sinfo) == IOTC_ER_NoERROR)
	{
		char *mode[3] = {"P2P", "RLY", "LAN"};
		// print session information(not a must)
		if( isdigit( Sinfo.RemoteIP[0] ))
			printf("Client is from[IP:%s, Port:%d] Mode[%s] VPG[%d:%d:%d] VER[%X] NAT[%d] AES[%d]\n", Sinfo.RemoteIP, Sinfo.RemotePort, mode[(int)Sinfo.Mode], Sinfo.VID, Sinfo.PID, Sinfo.GID, Sinfo.IOTCVersion, Sinfo.RemoteNatType, Sinfo.isSecure);
	}
	printf("avServStartEx OK, avIndex[%d], resend[%d] two_way_streaming[%d]\n", avIndex, avStartOutConfig.resend, avStartOutConfig.two_way_streaming);
    printf("auth_type[%d] account_or_identity[%s]\n\n", avStartOutConfig.auth_type, avStartOutConfig.account_or_identity);
    regedit_client_stream_mode(SID, avStartOutConfig.two_way_streaming);

#if ENABLE_DASA
    InitDasaSetting(avIndex);
#else
    avServSetResendSize(avIndex, 1024);
#endif

	while(1)
	{
		ret = avRecvIOCtrl(avIndex, &ioType, (char *)ioCtrlBuf, AV_MAX_IOCTRL_DATA_SIZE, 1000);
		if(ret >= 0)
		{
			Handle_IOCTRL_Cmd(SID, avIndex, ioCtrlBuf, ioType);
		}
		else if(ret != AV_ER_TIMEOUT)
		{
			printf("avIndex[%d], avRecvIOCtrl error, code[%d]\n",avIndex, ret);
			break;
		}
	}

	unregedit_client_from_video(SID);
	unregedit_client_from_audio(SID);
	avServStop(avIndex);
	printf("SID[%d], avIndex[%d], thread_ForAVServerStart exit!!\n", SID, avIndex);


	IOTC_Session_Close(SID);
	gOnlineNum--;

	pthread_exit(0);
}

static unsigned int getTimeStamp()
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (tv.tv_sec*1000 + tv.tv_usec/1000);
}

static void *thread_Login(void *arg)
{
    int ret, sleep_count;
    DeviceLoginInput auth_option;

    printf("thread_Login start\n");

    while(gProcessRun){
        auth_option.cb = sizeof(DeviceLoginInput);
        auth_option.authentication_type = AUTHENTICATE_BY_KEY;
        memcpy(auth_option.auth_key, gIotcAuthkey, IOTC_AUTH_KEY_LENGTH);

        ret = IOTC_Device_LoginEx((char *)arg, &auth_option);
        printf("IOTC_Device_LoginEx() ret = %d\n", ret);
        if(ret == IOTC_ER_NoERROR) {
            break;
        } else {
            PrintErrHandling (ret);
            for(sleep_count = 0 ; sleep_count < 5 && gProcessRun ; sleep_count++)
                sleep(1);
        }
    }

    printf("thread_Login exit\n");
    pthread_exit(0);
}

static int StreamoutSendVideo(unsigned int timestamp, char* buf, int size, int frameType)
{
    //printf("Video timestamp[%d] frameType[%s] size[%d]\n", timestamp, frameType == IPC_FRAME_FLAG_IFRAME ? "I" : "P", size);
    int i = 0, avIndex = 0, bEnableVideo = 0, ret = 0, lock_ret = 0;
    FRAMEINFO_t frameInfo;
    int waitKeyFrame = 0;
    char *videoBuf = buf;
    int videoSize = size;
    int sendFrameOut = 0;

    // *** set Video Frame info here ***
    memset(&frameInfo, 0, sizeof(FRAMEINFO_t));
    frameInfo.codec_id = MEDIA_CODEC_VIDEO_H264;
    frameInfo.flags = frameType;
    frameInfo.timestamp = timestamp;
    frameInfo.onlineNum = gOnlineNum;

    for(i = 0 ; i < MAX_CLIENT_NUMBER; i++)
    {
        //get reader lock
        lock_ret = pthread_rwlock_rdlock(&gClientInfo[i].sLock);
        if(lock_ret)
            printf("Acquire SID %d rdlock error, ret = %d\n", i, lock_ret);

        avIndex = gClientInfo[i].avIndex;
        bEnableVideo = gClientInfo[i].bEnableVideo;
        waitKeyFrame = gClientInfo[i].waitKeyFrame;

        //release reader lock
        lock_ret = pthread_rwlock_unlock(&gClientInfo[i].sLock);
        if(lock_ret)
            printf("Acquire SID %d rdlock error, ret = %d\n", i, lock_ret);

        if(avIndex < 0 || bEnableVideo == 0){
            continue;
        }

#if ENABLE_DASA
        bool clean_buf = IsCleanBufferInProgress(i);
        if(clean_buf) {
            continue;
        }
#endif
        if(waitKeyFrame) {
            if(frameInfo.flags == IPC_FRAME_FLAG_PBFRAME){
                continue;
            } else {
                lock_ret = pthread_rwlock_wrlock(&gClientInfo[i].sLock);
                if(lock_ret)
                    printf("Acquire SID %d rwlock error, ret = %d\n", i, lock_ret);

                gClientInfo[i].waitKeyFrame = 0;
                //release lock
                lock_ret = pthread_rwlock_unlock(&gClientInfo[i].sLock);
                if(lock_ret)
                    printf("Release SID %d rwlock error, ret = %d\n", i, lock_ret);
            }
        }

		// Send Video Frame to av-idx and know how many time it takes
		ret = avSendFrameData(avIndex, videoBuf, videoSize, &frameInfo, sizeof(FRAMEINFO_t));

		if(ret == AV_ER_EXCEED_MAX_SIZE) // means data not write to queue, send too slow, I want to skip it
		{
			lock_ret = pthread_rwlock_wrlock(&gClientInfo[i].sLock);
            if(lock_ret)
                printf("Acquire SID %d rwlock error, ret = %d\n", i, lock_ret);

            gClientInfo[i].waitKeyFrame = 1;

            //release lock
            lock_ret = pthread_rwlock_unlock(&gClientInfo[i].sLock);
            if(lock_ret)
                printf("Release SID %d rwlock error, ret = %d\n", i, lock_ret);
			continue;
		}
		else if(ret == AV_ER_SESSION_CLOSE_BY_REMOTE)
		{
			printf("thread_VideoFrameData AV_ER_SESSION_CLOSE_BY_REMOTE SID[%d]\n", i);
			unregedit_client_from_video(i);
			continue;
		}
		else if(ret == AV_ER_REMOTE_TIMEOUT_DISCONNECT)
		{
			printf("thread_VideoFrameData AV_ER_REMOTE_TIMEOUT_DISCONNECT SID[%d]\n", i);
			unregedit_client_from_video(i);
			continue;
		}
		else if(ret == IOTC_ER_INVALID_SID)
		{
			printf("Session cant be used anymore\n");
			unregedit_client_from_video(i);
			continue;
		}
#if ENABLE_DASA
		else if(ret == AV_ER_DASA_CLEAN_BUFFER)
		{
			printf("SID[%d] avIndex[%d] need to do clean buffer\n", i, avIndex);
			SetCleanBufferTakeAction(i);
		}
#endif
		else if(ret < 0)
		{
			printf("avSendFrameData avIndex[%d] error[%d]\n", avIndex, ret);
		}
		else
		{
			sendFrameOut = 1;
		}
	}

    return sendFrameOut;
}

static int StreamoutSendAudio(unsigned int timestamp, char* buf, int size)
{
    //printf("Audio timestamp[%d] size[%d]\n", timestamp, size);
    int i = 0, ret = 0;
	FRAMEINFO_t frameInfo;

    // *** set audio frame info here ***
    memset(&frameInfo, 0, sizeof(FRAMEINFO_t));
    frameInfo.codec_id = AUDIO_CODEC;
    frameInfo.flags = (AUDIO_SAMPLE_8K << 2) | (AUDIO_DATABITS_16 << 1) | AUDIO_CHANNEL_MONO;
    frameInfo.timestamp = timestamp;
    frameInfo.onlineNum = gOnlineNum;

    for(i = 0 ; i < MAX_CLIENT_NUMBER; i++)
    {
		//get reader lock
		int lock_ret = pthread_rwlock_rdlock(&gClientInfo[i].sLock);
		if(lock_ret)
			printf("Acquire SID %d rdlock error, ret = %d\n", i, lock_ret);
		if(gClientInfo[i].avIndex < 0 || gClientInfo[i].bEnableAudio == 0)
		{
			//release reader lock
			lock_ret = pthread_rwlock_unlock(&gClientInfo[i].sLock);
			if(lock_ret)
				printf("Acquire SID %d rdlock error, ret = %d\n", i, lock_ret);
			continue;
		}

		// send audio data to av-idx
		ret = avSendAudioData(gClientInfo[i].avIndex, buf, size, &frameInfo, sizeof(FRAMEINFO_t));
		//release reader lock
		lock_ret = pthread_rwlock_unlock(&gClientInfo[i].sLock);
		if(lock_ret)
			printf("Acquire SID %d rdlock error, ret = %d\n", i, lock_ret);

		//printf("avIndex[%d] size[%d]\n", gClientInfo[i].avIndex, size);
		if(ret == AV_ER_SESSION_CLOSE_BY_REMOTE)
		{
			printf("thread_AudioFrameData: AV_ER_SESSION_CLOSE_BY_REMOTE\n");
			unregedit_client_from_audio(i);
		}
		else if(ret == AV_ER_REMOTE_TIMEOUT_DISCONNECT)
		{
			printf("thread_AudioFrameData: AV_ER_REMOTE_TIMEOUT_DISCONNECT\n");
			unregedit_client_from_audio(i);
		}
		else if(ret == IOTC_ER_INVALID_SID)
		{
			printf("Session cant be used anymore\n");
			unregedit_client_from_audio(i);
		}
		else if(ret < 0)
		{
			printf("avSendAudioData error[%d]\n", ret);
			unregedit_client_from_audio(i);
		}
	}

    return 0;
}

/********
Thread - Send Audio frames to all AV-idx
*/
static void *thread_AudioFrameData(void *arg)
{
	FILE *fp=NULL;
	char buf[AUDIO_BUF_SIZE];
	int frameRate = AUDIO_FPS;
	int sleepTick = 1000000/frameRate;

	fp = fopen(gAudioFn, "rb");
	if(fp == NULL)
	{
		printf("thread_AudioFrameData: Audio File \'%s\' open error!!\n", gAudioFn);
		printf("[Audio] is DISABLED!!\n");
		printf("thread_AudioFrameData: exit\n");
		pthread_exit(0);
	}

	printf("thread_AudioFrameData start OK\n");
	printf("[Audio] is ENABLED!!\n");

	while(gProcessRun)
	{
		int size = fread(buf, 1, AUDIO_FRAME_SIZE, fp);
		if(size <= 0)
		{
			printf("rewind audio\n");
			rewind(fp);
			continue;
		}

		StreamoutSendAudio(getTimeStamp(), buf, size);

		usleep(sleepTick);
	}

	fclose(fp);

	printf("[thread_AudioFrameData] exit\n");

	pthread_exit(0);
}

/********
Thread - Send Video frames to all AV-idx
*/
static void *thread_VideoFrameData(void *arg)
{
    float hF = 0.0, lF= 0.0, totalFps = 0;
    long takeSec = 0, takeUSec = 0, sendFrameRoundTick = 0;
    int fpsCnt = 0, round = 0, framerate = FPS, sleepTick = 1000000/framerate;
    int bSendFrameOut = 0, size = 0;
    char buf[VIDEO_BUF_SIZE];
    struct timeval tv, tv2;
    struct timeval tStart, tEnd;
    FILE *fp = NULL;
    FRAMEINFO_t frameInfo;

	fp = fopen(gVideoFn, "rb");
	if(fp == NULL){
		printf("thread_VideoFrameData: Video File \'%s\' open error!!\n", gVideoFn);
		printf("[Vidio] is DISABLED!!\n");
		printf("thread_VideoFrameData: exit\n");
		pthread_exit(0);
	}

	// input file only one I frame for test
	size = fread(buf, 1, VIDEO_BUF_SIZE, fp);
	fclose(fp);
	if(size <= 0){
		printf("thread_VideoFrameData: Video File \'%s\' read error!!\n", gVideoFn);
		printf("[Vidio] is DISABLED!!\n");
		printf("thread_VideoFrameData: exit\n");
		pthread_exit(0);
	}

	// *** set Video Frame info here ***
	memset(&frameInfo, 0, sizeof(FRAMEINFO_t));
	frameInfo.codec_id = MEDIA_CODEC_VIDEO_H264;
	frameInfo.flags = IPC_FRAME_FLAG_IFRAME;

	printf("thread_VideoFrameData start OK\n");
	printf("[Video] is ENABLED!!\n");

	while(gProcessRun)
	{
		frameInfo.timestamp = getTimeStamp();
		bSendFrameOut = 0;
		takeSec = 0, takeUSec = 0, sendFrameRoundTick = 0;

#if ENABLE_DASA
        GetDASASuggestFrameRate(&framerate);
        sleepTick = 1000000/framerate;
#endif

        if(fpsCnt == 0)
            gettimeofday(&tv, NULL);

        gettimeofday(&tStart, NULL);
        bSendFrameOut = StreamoutSendVideo(getTimeStamp(), buf, size, frameInfo.flags);
        gettimeofday(&tEnd, NULL);

        takeSec = tEnd.tv_sec-tStart.tv_sec, takeUSec = tEnd.tv_usec-tStart.tv_usec;
        if(takeUSec < 0) {
            takeSec--;
            takeUSec += 1000000;
        }
        sendFrameRoundTick += takeUSec;

        if(bSendFrameOut && fpsCnt++ >= framerate)
        {
            round++;
            gettimeofday(&tv2, NULL);
            long sec = tv2.tv_sec-tv.tv_sec, usec = tv2.tv_usec-tv.tv_usec;
            if(usec < 0)
            {
                sec--;
                usec += 1000000;
            }
            usec += (sec*1000000);

            long oneFrameUseTime = usec / fpsCnt;
            float fps = (float)1000000/oneFrameUseTime;
            if(fps > hF) hF = fps;
            if(lF == 0.0) lF = fps;
            else if(fps < lF) lF = fps;
            printf("[AVAPIs_Servr] Send Video FPS = %f R[%d]\n", fps, round);
            fpsCnt = 0;
            totalFps += fps;
        }

        // notice the frames sending time for more specific frame rate control
        if( sleepTick > sendFrameRoundTick )
            usleep(sleepTick-sendFrameRoundTick);
    }

    printf("[thread_VideoFrameData] exit High/Low [%f/%f] AVG[%f]\n", hF, lF, (float)totalFps/round);

    pthread_exit(0);
}

/********
Thread - Fake Push Notification for demo
*/
#define DEFAULT_TEST_PUSH_COUNT 1
static void *ThreadFakePushNotification(void *arg) {
    printf("%s start\n", __func__);

    int ret = 0;
    TUTKJsonObject *notification_obj = NULL;
    unsigned int push_count = 0;

    ret = TUTK_Json_Obj_Create_From_String("{\"event\":\"1\"}", &notification_obj);
    if (ret != NEBULA_ER_NoERROR) {
        printf("TUTK_Json_Obj_Create_From_String err[%d]\n", ret);
        printf("%s exit\n", __func__);
        return NULL;
    }

    sleep(3);

    while (gProcessRun) {
        if (gSettingsExist && push_count < DEFAULT_TEST_PUSH_COUNT) {
            ret = avServPushNotification(notification_obj, 10000);
            printf("avServPushNotification[%s], ret[%d]\n", TUTK_Json_Obj_To_String(notification_obj), ret);
            push_count++;
        }
        sleep(5);
    }
    TUTK_Json_Obj_Release(notification_obj);
    printf("%s exit\n", __func__);
    pthread_exit(0);
}


static void InitAVInfo()
{
	int i;
	for(i=0;i<MAX_CLIENT_NUMBER;i++)
	{
		memset(&gClientInfo[i], 0, sizeof(AV_Client));
		gClientInfo[i].avIndex = -1;
		gClientInfo[i].playBackCh = -1;
		pthread_rwlock_init(&(gClientInfo[i].sLock), NULL);
	}
}

static void DeInitAVInfo()
{
	int i;
	for(i=0;i<MAX_CLIENT_NUMBER;i++)
	{
        pthread_rwlock_destroy(&gClientInfo[i].sLock);
		memset(&gClientInfo[i], 0, sizeof(AV_Client));
		gClientInfo[i].avIndex = -1;
	}
}

static int CreateFrameDataStream()
{
    pthread_t ThreadVideoFrameData_ID;
    pthread_t ThreadAudioFrameData_ID;
    int ret = 0;

	if((ret = pthread_create(&ThreadVideoFrameData_ID, NULL, &thread_VideoFrameData, NULL)))
	{
		printf("pthread_create ret=%d\n", ret);
		return -1;
	}
	pthread_detach(ThreadVideoFrameData_ID);

	if((ret = pthread_create(&ThreadAudioFrameData_ID, NULL, &thread_AudioFrameData, NULL)))
	{
		printf("pthread_create ret=%d\n", ret);
		return -1;
	}
	pthread_detach(ThreadAudioFrameData_ID);
    
    return 0;
}

static void create_streamout_thread()
{
    int ret = 0;

#if ENABLE_EMULATOR
// Enable the emulator stream for sample use only.
// Users can exclude the code when doing integration.
    ret = CreateEmulatorStream(StreamoutSendVideo, StreamoutSendAudio);
#else
    ret = CreateFrameDataStream();
#endif
    if(ret != 0){
        printf("create stream thread failed ret=%d\n", ret);
        exit(-1);
    }

#if ENABLE_DASA
	ret = CreateDasaCheckThread();
    if(ret != 0) {
        printf("create DASA check thread failed ret=%d\n", ret);
        exit(-1);
    }
#endif
}

static void LoginInfoCB(unsigned int nLoginInfo)
{
	if((nLoginInfo & 0x04))
		printf("I can be connected via Internet\n");
}

static void PrintUsage()
{
    printf("#########################################################################\n");
    printf("./AVAPIs_Server [options...]\n");
    printf("Options: (M)Must (O)Option\n");
    printf(" -u [UID]\t\t(M)Config UID.\n");
    printf(" -v [video file]\t(O)Config video file path for demo. Default path is video/I_000.bin\n");
    printf(" -a [audio file]\t(O)Config audio file path for demo. Default path is audio/8k_16bit_mono.raw\n");
    printf(" -f [ability file]\t(O)Config ability file path for demo.\n");
    printf(" -p \t\t\t(O)Enable push notification. Default path of notification settings file is ./notification_setting.txt\n");
    printf(" -h \t\t\t   Show usage\n");
}

static int ParseInputOptions(int argc, char *argv[])
{
    int option = 0;
    while((option = getopt(argc, argv, "u:a:v:f:hp")) > 0){
        switch(option) {
          case 'u':
            strncpy(gUID, optarg, sizeof(gUID));
            gUID[MAX_UID_LENGTH] = '\0';
            printf("gUID %s\n", gUID);
            break;
          case 'a':
            strncpy(gAudioFn, optarg, sizeof(gAudioFn));
            printf("Audio file %s\n", gAudioFn);
            break;
          case 'v':
            strncpy(gVideoFn, optarg, sizeof(gVideoFn));
            printf("Video file %s\n", gVideoFn);
            break;
          case 'f':
            strncpy(gAbilityFile, optarg, sizeof(gAbilityFile));
            printf("Ability file %s\n", gAbilityFile);
            break;
          case 'p':
            gEnablePushNotification = true;
            break;
          case 'h':
            return -1;
          default:
            printf("Unknown option %c\n", option);
            return -1;
        }
    }
    if(strlen(gUID) == 0){
        return -1;
    }
    return 0;
}


int main(int argc, char *argv[])
{
	int ret , SID;
	pthread_t ThreadLogin_ID;

    ret = ParseInputOptions(argc, argv);
    if(ret < 0){
        printf("ParseInputOptions ret[%d]\n", ret);
        PrintUsage();
        return -1;
    }

    const char *license_key = getenv(ENV_LICENSE_KEY);
    if (license_key == NULL) {
        printf("Environment variable %s must be set!!\n", ENV_LICENSE_KEY);
        return -1;
    }

	printAccount();

	//set the output path of the log file when using log library.
    LogAttr log_attr;
    memset(&log_attr, 0, sizeof(log_attr));

    log_attr.path = "./log_iotc_device.txt";
    ret = IOTC_Set_Log_Attr(log_attr);
    printf("IOTC_Set_Log_Attr ret[%d]\n", ret);

    log_attr.path = "./log_global_device.txt";
    ret = TUTK_Set_Log_Attr(log_attr);
    printf("TUTK_Set_Log_Attr ret[%d]\n", ret);

    log_attr.path = "./log_av_device.txt";
    ret = AV_Set_Log_Attr(log_attr);
    printf("AV_Set_Log_Attr ret[%d]\n", ret);

    do {
        InitAVInfo();
#if ENABLE_TOKEN_AUTH
        ret = LoadTokenFile("extension/token/token.txt");
#endif
        if(ret<0)
            printf("load token file failed\n");

        create_streamout_thread();

        ret = TUTK_SDK_Set_License_Key(license_key);
        if (ret != TUTK_ER_NoERROR)
        {
            printf("TUTK_SDK_Set_License_Key()=%d ,exit...!!\n", ret);
            return 0;
        }

        IOTC_Set_Max_Session_Number(MAX_CLIENT_NUMBER);
        // use which Master base on location, port 0 means to get a random port
        ret = IOTC_Initialize2(0);
        if(ret != IOTC_ER_NoERROR)
        {
            printf("  [] IOTC_Initialize2(), ret=[%d]\n", ret);
            PrintErrHandling (ret);
            DeInitAVInfo();
            return -1;
        }

        printf("IOTCAPI version[%s] AVAPI version[%s]\n", IOTC_Get_Version_String(), avGetAVApiVersionString());

        IOTC_Get_Login_Info_ByCallBackFn(LoginInfoCB);
        // alloc MAX_CLIENT_NUMBER*3 for every session av data/speaker/play back
        avInitialize(MAX_CLIENT_NUMBER*3);

        // create thread to login because without WAN still can work on LAN
        if((ret = pthread_create(&ThreadLogin_ID, NULL, &thread_Login, (void *)gUID)))
        {
            printf("Login Thread create fail, ret=[%d]\n", ret);
            return -1;
        }
        pthread_detach(ThreadLogin_ID);
        
        if(gEnablePushNotification) {
            char *settings = NULL;
            FILE *fp = NULL;
            // get notification settings
            fp = fopen(gDefaultSettingsFilePath, "r");
            if (fp != NULL) {
                fseek(fp, 0, SEEK_END);
                long fsize = ftell(fp);
                fseek(fp, 0, SEEK_SET);
                settings = calloc(1, fsize + 1);
                ret = fread(settings, 1, fsize, fp);
                fclose(fp);
                if(ret <= 0){
                    printf("read %s fail, ret=[%d]\n", gDefaultSettingsFilePath, ret);
                    return -1;
                }
                gSettingsExist = true;
            } else {
                printf("%s not exist\n", gDefaultSettingsFilePath);
            }

            ret = avServEnablePushNotification(gUID, settings, NotificationSettingHandleFn);
            printf("avServEnablePushNotification ret[%d]\n", ret);

            if(settings)
                free(settings);

            pthread_t fake_push_thread_id;
            if (pthread_create(&fake_push_thread_id, NULL, ThreadFakePushNotification, NULL) < 0) {
                printf("create ThreadFakePushNotification failed!\n");
                return -1;
            }
            pthread_detach(fake_push_thread_id);
        }

        gProcessRun = 1;
        while(gProcessRun)
        {
            // Accept connection only when IOTC_Listen() calling
            SID = IOTC_Listen(1000);
            if (SID < 0) {
                PrintErrHandling (SID);
                if (SID == IOTC_ER_EXCEED_MAX_SESSION) {
                    sleep(5);
                }
                continue;
            }

            int *sid = (int *)malloc(sizeof(int));
            *sid = SID;
            pthread_t Thread_ID;
            ret = pthread_create(&Thread_ID, NULL, &thread_ForAVServerStart, (void *)sid);
            if(ret < 0) {
                printf("pthread_create failed ret[%d]\n", ret);
            } else {
                pthread_detach(Thread_ID);
                gOnlineNum++;
            }

        }
        DeInitAVInfo();
        avDeInitialize();
        IOTC_DeInitialize();
        break;

    } while(1);

	return 0;
}
